//
// Copyright (c) 1996 Ben Ross
//
// You may distribute under the terms of the GNU General Public
// License as specified in the COPYING file.
//

// $Id: XeApp.h,v 1.9 2001/08/04 11:27:56 ben Exp $

#ifndef _XeApp_H_
#define _XeApp_H_

#include <Xe.h>
#include <X11/Xlib.h>
#include <X11/Xresource.h>

#include <Kr/KrCB.h>
#include <Kr/KrTime.h>
#include <Kr/KrAtom.h>

typedef int 	(*XErrorHandler)(Display*, XErrorEvent*);

class XeObject;
class XeTNode;

class XeApp {
public:
    XeApp(const char* className, int& argc, char** argv);
    ~XeApp(void);
    
    void    run(void);
    
    void	setTimeout(const KrCB& cb, KrTime time);
    void	clearTimeout(void);
    
    
    // Temporarily here so XeText can use it.
    void	doUpdate(void);
    
    bool	setRunOnce(const KrCB& cb, int argc, char** argv);

    struct RunOnceCBData {
    	const char*	cwd;
    	int			argc;
    	char**		argv;
    };
    
private:	
	// number of seconds before app times out (quits)
	long		_timeout;	
	XeTNode*	_tout;
	
	KrCB		_runOnceCB;
	
	Atom		_windowIdRoot;	// XA_WINDOW
	Atom		_windowArg;		// XA_STRING
	Atom		_windowUid;		// XA_INTEGER
	Atom		_windowIPAddr;	// XA_INTEGER

    void	doEvent(XEvent* theEvent);

    void	mergeResourceDatabases(void);    

};

//
// Timeout Class
//

class XeTNode {
public:
//	void*	getAuxData() { return _aux; }
	KrTime	getPeriod() { return _period; }
	KrTime	getTime() { return _when; }
	
private:
	XeTNode(const KrCB& cb, KrTime time)
		: _cb(cb), _period(time) {
			_when = KrTime::now();
		}
	
	void	call(void) {
			_when += _period;
			_cb(this);
		}
	
	KrCB		_cb;
	KrTime		_period;
	KrTime		_when;
	
	friend class XeApp;
};

// Globals:

extern XeApp*			gApp;

extern const char		*gProgname;
extern const char		*gClassname;
extern KrAtom			gCwd;
extern XrmQuark			gProgQuark;
extern XrmQuark			gClassQuark;

extern Display	*gDisplay;
extern Screen	*gScreenPtr;
extern int		gScreenNum;

extern int		gDepth;
extern int		gVisualClass;
extern Visual*	gVisual;

class XeDialog;

#include <ext/slist>

extern std::slist<XeDialog*>	gDialogListUp;
extern std::slist<XeDialog*>	gDialogList;

extern Atom		XA_WM_PROTOCOLS;

extern int		gGeometryFlags;
extern int		gX, gY;
extern uint		gWidth, gHeight;

//
// Functions:
//

int portaselect(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
    struct timeval *timeout);

bool getAppResource(const char* res, const char* clas, XrmValue* val);

#endif
